--
-- Metadata for generating data entry forms
--
-- Based on CMS form metadata code.
--
-- The core datamodel allows applications to store metadata to generate html
-- forms for use with dynamic object types.

---------------------------------------------
-- Form Layout definition
---------------------------------------------

create or replace function dtype_layout__create (integer,varchar,varchar,varchar,varchar,varchar,varchar)
returns integer as '
declare
  p_layout_id           alias for $1;  
  p_name                alias for $2;  
  p_pretty_name         alias for $3;  
  p_description         alias for $4;
  p_type                alias for $5;  
  p_filename            alias for $6;
  p_resource_dir        alias for $7;
  v_layout_id           dtype_layouts.layout_id%TYPE;
begin

    select coalesce(p_layout_id, nextval(''t_dtype_seq'')) into v_layout_id;
  
    insert into dtype_layouts (
        layout_id,
        type,
        name,
        pretty_name,
        description,
        filename,
        resource_dir
    )
    values (
        v_layout_id,
        p_type,
        p_name,
        p_pretty_name,
        p_description,
        p_filename,
        p_resource_dir
    );

    return v_layout_id; 
end;' language 'plpgsql';

create or replace function dtype_layout__delete (integer)
returns integer as '
declare
  p_layout_id           alias for $1;  
begin
  
    delete from dtype_layouts where layout_id = v_layout_id;

    return 0; 
end;' language 'plpgsql';


---------------------------------------------
-- Form Package definition
---------------------------------------------

create or replace function dtype_form__create (varchar,varchar,varchar)
returns integer as '
declare
  p_layout              alias for $1;  
  p_widget              alias for $2;
  p_custom              alias for $3;
  v_form_id             dtype_forms.form_id%TYPE;
  v_layout_id           dtype_layouts.layout_id%TYPE;
  v_element_id          dtype_elements.element_id%TYPE;
begin
    select nextval(''t_dtype_seq'') into v_form_id;

    insert into dtype_forms (
        form_id
    )
    values (
        v_form_id
    );

    select nextval(''t_dtype_seq'') into v_element_id;

    -- Look for the layout
    select layout_id into v_layout_id
      from dtype_layouts
     where type = ''form''
       and name = p_layout;

    -- Create the root section
    insert into dtype_elements (
        element_id,
        name,
        form_id,
        parent_id,
        layout_id,
        widget,
        custom
    )
    values (
        v_element_id,
        ''root'',
        v_form_id,
        null,
        v_layout_id,
        p_widget,
        p_custom
    );

    return v_form_id; 
end;' language 'plpgsql';

create or replace function dtype_form__create (varchar)
returns integer as '
declare
  p_layout              alias for $1;  
begin
    return dtype_form__create (p_layout, null, null);
end;' language 'plpgsql';

create or replace function dtype_form__delete (integer)
returns integer as '
declare
  p_form_id             alias for $1;  
  v_form_id             dtype_forms.form_id%TYPE;
  v_element_id          dtype_elements.element_id%TYPE;
begin
    -- Look for the form
    select form_id into v_form_id
      from dtype_forms
     where form_id = p_form_id;

    -- if it is already deleted
    if v_form_id is null then 
        return 0;
    end if;

    -- Look for the root element
    select element_id into v_element_id
      from dtype_elements
     where parent_id is null
       and form_id = v_form_id;

    if v_element_id is not null then
        perform dtype_element__delete(v_element_id);
    end if;

    delete from dtype_forms where form_id = v_form_id;
  
    return 0; 
end;' language 'plpgsql';

create or replace function dtype_form__map_type (integer,varchar,varchar)
returns integer as '
declare
  p_form_id             alias for $1;  
  p_form                alias for $2;  
  p_object_type         alias for $3;  
  v_form_id             dtype_forms.form_id%TYPE;
  v_object_type         dtype_form_type_map.object_type%TYPE;
begin
    select form_id into v_form_id
      from dtype_form_type_map
     where form_id = p_form_id
       and object_type = p_object_type;

    if v_form_id is null then
        insert into dtype_form_type_map (
            form_id,
            form,
            object_type
        )
        values (
            p_form_id,
            p_form,
            p_object_type
        );
    else
        update dtype_form_type_map 
           set form = p_form
         where form_id = p_form_id
           and object_type = p_object_type;
    end if;

    return 0;
end;' language 'plpgsql';


---------------------------------------------
-- Element package definition
---------------------------------------------

create or replace function dtype_element__create (integer,varchar,integer,varchar,varchar,varchar)
returns integer as '
declare
  p_element_id          alias for $1;  
  p_name                alias for $2;  
  p_parent_id           alias for $3;
  p_widget              alias for $4;
  p_layout              alias for $5;  
  p_custom              alias for $6;  
  v_element_id          dtype_elements.element_id%TYPE;
  v_form_id             dtype_elements.form_id%TYPE;
  v_layout_id           dtype_elements.layout_id%TYPE;
  param_rec             record;
begin
    select coalesce(p_element_id, nextval(''t_dtype_seq'')) into v_element_id;
  
    select form_id into v_form_id
      from dtype_elements
     where element_id = p_parent_id;

    -- Look for the layout
    select layout_id into v_layout_id
      from dtype_layouts
     where type = ''element''
       and name = p_layout;

    insert into dtype_elements (
        element_id,
        name,
        form_id,
        parent_id,
        widget,
        layout_id,
        custom
    )
    values (
        v_element_id,
        p_name,
        v_form_id,
        p_parent_id,
        p_widget,
        v_layout_id,
        p_custom
    );

    -- create required params with default values
    for param_rec in select param_id, default_value 
                       from dtype_proc_params 
                      where proc = ''template::element::create''
                        and is_required = ''t''
    loop
        perform dtype_proc_param_value__create (
            null,                    -- value_id
            v_element_id,            -- call_id
            param_rec.param_id,      -- param_id
            ''literal'',             -- param_source
            param_rec.default_value  -- value
        );
    end loop;


    return v_element_id; 
end;' language 'plpgsql';

create or replace function dtype_element__create (integer,varchar,integer,varchar,varchar)
returns integer as '
declare
  p_element_id          alias for $1;  
  p_name                alias for $2;  
  p_parent_id           alias for $3;
  p_widget              alias for $4;
  p_layout              alias for $5;  
begin
    return dtype_element__create(p_element_id, p_name, p_parent_id, p_widget, p_layout); 
end;' language 'plpgsql';

create or replace function dtype_element__delete (integer)
returns integer as '
declare
  p_element_id           alias for $1;  
  v_form_id              dtype_elements.form_id%TYPE;
  v_sortkey              dtype_elements.tree_sortkey%TYPE;
  v_element              RECORD;
begin
    select form_id, tree_sortkey
    into v_form_id, v_sortkey
    from dtype_elements
    where element_id = p_element_id;

    -- if it is already deleted
    if v_form_id is null
    then return 0;
    end if;

    -- delete all children
    -- order by tree_sortkey desc to guarantee
    -- that we never delete a parent before its child
    -- sortkeys are beautiful
    for v_element in select *
                     from dtype_elements
                     where form_id = v_form_id
                     and tree_sortkey >= tree_left(v_sortkey) 
                     and tree_sortkey < tree_right(v_sortkey)
                     order by tree_sortkey desc
    loop
        delete from dtype_elements where element_id = v_element.element_id;
    end loop;

    -- that deleted all the children. now delete the element
    delete from dtype_elements where element_id = p_element_id;

    return 0; 
end;' language 'plpgsql';

create or replace function dtype_element__map_attribute (integer,integer)
returns integer as '
declare
  p_element_id          alias for $1;  
  p_attribute_id        alias for $2;  
  v_element_id          dtype_elements.element_id%TYPE;
begin
    select element_id into v_element_id
      from dtype_element_attribute_map
     where element_id = p_element_id
       and attribute_id = p_attribute_id;

    if v_element_id is null then
        insert into dtype_element_attribute_map (
            element_id,
            attribute_id
        )
        values (
            p_element_id,
            p_attribute_id
        );
    end if;

    return 0;
end;' language 'plpgsql';

--
-- A trigger to update the form elements sort key.
--
-- This code is lifted from ben@openforce.biz forums code.
--

create function dtype_element_insert_tr ()
returns opaque as '
declare
    v_max_child_sortkey             dtype_elements.max_child_sortkey%TYPE;
    v_parent_sortkey                dtype_elements.tree_sortkey%TYPE;
begin

    select tree_sortkey, max_child_sortkey
    into v_parent_sortkey, v_max_child_sortkey
    from dtype_elements
    where element_id = new.parent_id
    for update;

    v_max_child_sortkey := tree_increment_key(v_max_child_sortkey);

    update dtype_elements
    set max_child_sortkey = v_max_child_sortkey
    where element_id = new.parent_id;

    new.tree_sortkey := coalesce(v_parent_sortkey, '''') || v_max_child_sortkey;

    return new;
end;' language 'plpgsql';

create trigger dtype_element_insert_tr
before insert on dtype_elements
for each row
execute procedure dtype_element_insert_tr();

