--
-- Metadata for invoking parameter callbacks.
--
-- Allows the values for some or all of the arguments to a proc to be 
-- stored in the database.  This means that some or all of the behaviour of a
-- proc can be defined in a template, in an instance or at the point where the 
-- proc is being called.
--
-- This might be the craziest thing I have ever written... 
--
-- @author Lee Denison (lee@xarg.co.uk)

---------------------------------------------
-- Core reference datamodel
---------------------------------------------

create table dtype_procs (
  name             varchar(1000)
                   constraint dtype_procs_pk
                   primary key
);

comment on table dtype_procs is '
  Canonical list of all metadata driven procs for use with dtypes.
';

create table dtype_proc_params (
  param_id         integer
                   constraint dtype_proc_params_pk
                   primary key,
  proc             varchar(100)
                   constraint dtype_proc_params_fk
                   references dtype_procs,
  name             varchar(100)
                   constraint dtype_params_name_nil
                   not null,
  param_type       varchar(100) default 'switch:onevalue'
                   constraint dtype_proc_param_type_nil
                   not null
                   constraint dtype_proc_param_type_ck
                   check (param_type in ('switch:onevalue', 'switch:boolean', 'switch:spec')),
  is_required      boolean,
  default_value    text,
  constraint dtype_proc_name_unq unique (proc, name)
);


comment on table dtype_proc_params is '
  Parameters that are specific to a particular proc.
';

comment on column dtype_proc_params.param_type is '
  Determines how the dtype:: api will use this parameter when invoking the 
  proc:

    switch:onevalue  - will be added to the call as a -<param> <value> switch.
    
    switch:boolean   - will be added to the call as a -<param> flag if the 
                       value is 1, otherwise nothing will be added to the call.

    switch:spec      - will be added to the call as a switch of the form:

                           -<param> { <subparam> <value> } 

                       Where param and subparam are stored in the name field as 

                           <param>,<subparam> 

                       All spec params with the same major part will be grouped
                       into a single spec.

  New types may be added in future to further control behaviour.
';

create table dtype_proc_param_values (
  value_id            integer
                      constraint dtype_proc_param_val_pk
                      primary key,
  call_id             integer
                      constraint dtype_proc_param_val_call_nn
                      not null,
  param_id            integer
                      constraint dtype_proc_param_id_fk
                      references dtype_proc_params
                      constraint dtype_proc_param_id_nn
                      not null,
  param_source        varchar(100) default 'literal'
                      constraint dtype_proc_param_val_src_nil
                      not null
                      constraint dtype_proc_param_val_src_ck 
                      check (param_source in ('literal', 'scope', 'proc')),
  value		          text,
  constraint dtype_proc_param_val_unq
  unique (call_id, param_id)
);

comment on table dtype_proc_param_values is '
  Sets of values that should be used together.
';

comment on column dtype_proc_param_values.param_source is '
  ''literal''   - corresponds to a literal value stored in the value column.
  ''scope''     - corresponds to a variable expected to be in scope when the
                  proc is evaluated.
  ''proc''      - corresponds to a proc name and value set that should be 
                  invoked to retrieve the parameter value.  Stored as a space
                  separated pair.
';
