--
-- Form and element templates datamodel.
--
-- Element templates are provided to allow common widget / datatype combinations
-- to be named and offered in user interfaces that create object types.  
-- Default widgets identify which widget templates to use for default object
-- type forms.

---------------------------------------------
-- Element templates datamodel
---------------------------------------------

create table dtype_element_tmpls (
  template_id      integer
                   constraint dtype_elm_tmpl_tid_fk
                   references acs_objects(object_id)
                   constraint dtype_elm_tmpl_tid_pk
                   primary key,
  name             varchar(100)
                   constraint dtype_elm_tmpl_unq
                   unique,
  pretty_name      varchar(1000)
                   constraint dtype_elm_tmpl_pname_nn
                   not null,
  datatype         varchar(50)
                   constraint dtype_elm_tmpl_dtype_fk
                   references acs_datatypes,
  widget           varchar(100)
                   constraint dtype_elm_tmpl_wgt_fk
                   references dtype_widgets
);

comment on table dtype_element_tmpls is '
  A list of common datatype / widget combinations and what to call them.
';

select acs_object_type__create_type (
    'dtype_element_template',       -- object_type
    'DType Element Template',       -- pretty_name
    'DType Element Templates',      -- pretty_plural
    'acs_object',                   -- supertype
    'dtype_element_tmpls',          -- table_name
    'template_id',                  -- id_column
    null,                           -- package_name
    'f',                            -- abstract_p
    null,                           -- type_extension_table
    'dtype_element_tmpl__title'     -- name_method
);

select acs_attribute__create_attribute (
   'dtype_element_template',        -- object_type
   'name',                          -- attribute_name
   'text',                          -- datatype
   'Short Name',                    -- pretty_name
   'Short Names',                   -- pretty_plural
   null,                            -- table_name
   null,                            -- column_name
   null,                            -- default_value
   1,                               -- min_n_values
   1,                               -- max_n_values
   2,                               -- sort_order
   'type_specific',                 -- storage
   'f'                              -- static_p
);

select acs_attribute__create_attribute (
   'dtype_element_template',        -- object_type
   'pretty_name',                   -- attribute_name
   'text',                          -- datatype
   'Name',                          -- pretty_name
   'Names',                         -- pretty_plural
   null,                            -- table_name
   null,                            -- column_name
   null,                            -- default_value
   1,                               -- min_n_values
   1,                               -- max_n_values
   1,                               -- sort_order
   'type_specific',                 -- storage
   'f'                              -- static_p
);

select acs_attribute__create_attribute (
   'dtype_element_template',        -- object_type
   'datatype',                      -- attribute_name
   'text',                          -- datatype
   'Datatype',                      -- pretty_name
   'Datatypes',                     -- pretty_plural
   null,                            -- table_name
   null,                            -- column_name
   'text',                          -- default_value
   1,                               -- min_n_values
   1,                               -- max_n_values
   3,                               -- sort_order
   'type_specific',                 -- storage
   'f'                              -- static_p
);

select acs_attribute__create_attribute (
   'dtype_element_template',        -- object_type
   'widget',                        -- attribute_name
   'text',                          -- datatype
   'Widget',                        -- pretty_name
   'Widgets',                       -- pretty_plural
   null,                            -- table_name
   null,                            -- column_name
   'text',                          -- default_value
   1,                               -- min_n_values
   1,                               -- max_n_values
   4,                               -- sort_order
   'type_specific',                 -- storage
   'f'                              -- static_p
);

---------------------------------------------
-- Package definition
---------------------------------------------

create or replace function dtype_element_tmpl__new (integer,integer,varchar,varchar,varchar,varchar,integer,varchar)
returns integer as '
declare
  p_template_id         alias for $1;  
  p_package_id          alias for $2;
  p_name                alias for $3;  
  p_pretty_name         alias for $4;  
  p_widget              alias for $5;  
  p_datatype            alias for $6;  
  p_creation_user       alias for $7;
  p_creation_ip         alias for $8;
  v_template_id         integer;
begin
  
    v_template_id := acs_object__new (
        p_template_id,
        ''dtype_element_tmpl'',
        current_timestamp,
        p_creation_user,
        p_creation_ip,
        p_package_id
    );

    insert into dtype_element_tmpls (
        template_id,
        name,
        pretty_name,
        datatype,
        widget
    ) values (
        v_template_id,
        p_name,
        p_pretty_name,
        p_datatype,
        p_widget
    );

    return v_template_id; 
end;' language 'plpgsql';

create or replace function dtype_element_tmpl__delete (integer)
returns integer as '
declare
    p_template_id alias for $1;
begin
    -- delete parameters associated with this widget template
    delete from dtype_element_tmpl_params
    where template_id = p_template_id;

    delete from dtype_element_tmpls
    where template_id = p_template_id;
    PERFORM acs_object__delete(p_template_id);
    return 0;
end;
' language 'plpgsql';

create or replace function dtype_element_tmpl__set_param_value (varchar,varchar,varchar,varchar,varchar,varchar,varchar)
returns integer as '
declare
  p_template_name       alias for $1;  
  p_param               alias for $2;  
  p_value               alias for $3;  
  p_param_type          alias for $4;  -- default ''one_value''
  p_param_source        alias for $5;  -- default ''literal''
  v_template_id         dtype_element_tmpls.template_id%TYPE;
  v_param_id            dtype_widget_params.param_id%TYPE;
  v_widget              dtype_widgets.widget%TYPE;
  v_prev_value          integer;       
begin

    -- Get the template id and the widget 
    select wt.template_id, 
           wt.widget 
      into v_template_id, 
           v_widget 
      from widget_templates wt
     where wt.name = p_template_name;

    if NOT FOUND then
      raise EXCEPTION ''-20000: No widget template exists with the name % in dtype_element_tmpl__set_param_value'', p_template_name;
    end if;

    -- Get the param id
    select param_id into v_param_id from dtype_widget_params
     where widget = v_widget 
       and param = p_param;

    if NOT FOUND then
      raise EXCEPTION ''-20000: No parameter named % exists for the widget % in dtype_widget__set_param_value'', p_param, v_widget;
    end if;  

    -- Check if an old value exists
    -- Determine if a previous value exists
    select count(1) into v_prev_value from dual 
      where exists (select 1 from dtype_element_tmpl_params
                    where template_id = v_template_id
                    and param_id = v_param_id);
    
    if v_prev_value > 0 then
      -- Update the value
      update dtype_element_tmpl_params set
        param_type = p_param_type,
        param_source = p_param_source,
        value = p_value
      where
        template_id = v_template_id
      and
        param_id = v_param_id;
    else
      -- Insert a new value
      insert into dtype_element_tmpl_params
        (template_id, param_id, param_type, param_source, value)
      values
        (v_template_id, v_param_id, p_param_type, p_param_source, p_value);
    end if;

    return 0; 
end;' language 'plpgsql';
